@@ -2,18 +2,11 @@
 
 import pandas as pd
 import numpy as np
-from matplotlib import pyplot as plt
-import seaborn as sns
+from typing import Iterable
 from pathlib import Path
 import json
-from typing import Callable
-from operator import index
 from pathlib import Path
 import sys
-import streamlit as st
-import pandas as pd
-import seaborn as sns
-from matplotlib import pyplot as plt
 
 __all__ = ["has_visdata", "main", "VISDATA_FOLDER"]
 
@@ -41,6 +34,8 @@ PATHMAP = dict()
 
 
 def adjust_type_names(df):
+    if df is None or len(df) == 0:
+        return df
     for c in df.columns:
         if not c.endswith("type"):
             continue
@@ -80,7 +75,7 @@ def parse_tournament(path, t_indx, base_indx):
     configs = json.load(open(path / CONFIGS_FILE))
     if not configs:
         return None, None, None
-    scores = pd.read_csv(path / SCORES_FILE).to_dict(orient="records")
+    scores = pd.read_csv(path / SCORES_FILE).to_dict("records") # typing: none
     if not scores:
         return None, None, None
     worlds = []
@@ -156,9 +151,203 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
             "dropped_at",
         ],
     )
+    negotiations = pd.DataFrame(
+        data=[],
+        columns=[
+            "id",
+            "seller",
+            "buyer",
+            "seller_type",
+            "buyer_type",
+            "delivery_step",
+            "delivery_relative_time",
+            "quantity",
+            "unit_price",
+            "step",
+            "relative_time",
+            "timedout",
+            "broken",
+            "issues",
+            "caller",
+            "product",
+            "rounds",
+            "world",
+            "tournament",
+            "min_quantity",
+            "max_quantity",
+            "min_delivery_step",
+            "max_delivery_step",
+            "min_delivery_relative_time",
+            "max_delivery_relative_time",
+            "min_unit_price",
+            "max_unit_price",
+        ],
+    )
+
+    offers_list = []
+    offers = pd.DataFrame(
+        data=[],
+        columns=[
+            "sender",
+            "receiver",
+            "sender_type",
+            "receiver_type",
+            "delivery_step",
+            "quantity",
+            "unit_price",
+            "step",
+            "negotiation",
+            "relative_time",
+            "timedout",
+            "broken",
+            "caller",
+            "product",
+            "world",
+            "tournament",
+            "round",
+        ],
+    )
     for cname in CONTRACTS_FILE:
         if nonzero(path / cname):
             contracts = pd.read_csv(path / cname, index_col=0)
+    if nonzero(path / NEGOTIATIONS_FILE):
+        negotiations = pd.read_csv(path / NEGOTIATIONS_FILE, index_col=0)
+        negotiations = negotiations.loc[
+            :,
+            [
+                "id",
+                "agreement",
+                "step",
+                "timedout",
+                "broken",
+                "issues",
+                "buyer",
+                "seller",
+                "caller",
+                "product",
+                "step",
+                "offers",
+                "ended_at",
+            ],
+        ]
+        negotiations = negotiations.rename(columns=dict(step="rounds"))
+        negotiations = negotiations.rename(columns=dict(ended_at="step"))
+        negotiations["offers"] = negotiations.offers.apply(lambda x: eval(x) if isinstance(x, str) else x)
+        negotiations["rounds"] = negotiations["rounds"] - 1
+        for indx, neg in negotiations.iterrows():
+            d = dict(
+                sender=None,
+                receiver=None,
+                delivery_step=None,
+                quantity=None,
+                unit_price=None,
+                round=None,
+                index=None,
+                step=neg["step"],
+                timedout=neg["timedout"],
+                broken=neg["broken"],
+                caller=neg["caller"],
+                product=neg["product"],
+                world=wname,
+                tournament=tname,
+            )
+            if not neg["offers"]:
+                continue
+            offers = eval(neg["offers"]) if isinstance(neg["offers"], str) else neg["offers"]
+            current_offers = []
+            for k, vs in offers.items():
+                for r, v in enumerate(vs):
+                    dd = {**d}
+                    dd["sender"] = k
+                    dd["negotiation"] = neg["id"]
+                    dd["receiver"] = neg["seller"] if k == neg["buyer"] else neg["buyer"]
+                    dd["round"] = r
+                    if v:
+                        dd["quantity"] = v[0]
+                        dd["delivery_step"] = v[1]
+                        dd["unit_price"] = v[2]
+                    else:
+                        dd["quantity"] = None
+                        dd["delivery_step"] = None
+                        dd["unit_price"] = None
+
+                    current_offers.append(dd)
+            for indx, v in enumerate(current_offers):
+                v["index"] = indx
+            offers_list += current_offers
+
+        atmap = dict(zip(agents["name"].to_list(), agents["type"].to_list()))
+        # negotiations.agreement[negotiations.agreement.isna(), "agreement"] = None
+        for c in ["quantity", "delivery_step", "unit_price"]:
+            negotiations[c] = None
+
+        def lst(x):
+            return (
+                [-1, -1, -1]
+                if not x or (isinstance(x, str) and x.lower() == "none")
+                else list(eval(x))
+                if isinstance(x, str)
+                else list(x)
+            )
+
+        negotiations.agreement = negotiations.agreement.apply(lst)
+        agreements = pd.DataFrame(negotiations.agreement.tolist(), index=negotiations.index)
+        agreements.columns = ["quantity", "delivery_step", "unit_price"]
+        for c in ["quantity", "delivery_step", "unit_price"]:
+            negotiations[c] = agreements[c]
+            negotiations.loc[negotiations.loc[:, c] < 0, c] = float("nan")
+
+        def do_map(x):
+            return atmap[x]
+
+        negotiations["buyer_type"] = negotiations.buyer.apply(do_map)
+        negotiations["seller_type"] = negotiations.seller.apply(do_map)
+        negotiations["world"] = wname
+        negotiations["tournament"] = tname
+        negotiations = negotiations.drop("agreement", axis=1)
+
+        def get_issue_ranges(x):
+            if isinstance(x, str):
+                x = eval(x)
+            if not isinstance(x, Iterable):
+                return x
+            y = []
+            for s in x:
+                mn, mx = eval(s.split(":")[-1].strip())
+                y += [mn, mx]
+            return y
+
+        negotiations = negotiations.reset_index(None)
+        issues = negotiations.issues.apply(get_issue_ranges)
+        issues = pd.DataFrame(
+            issues.to_list(),
+            columns=[
+                "min_quantity",
+                "max_quantity",
+                "min_delivery_step",
+                "max_delivery_step",
+                "min_unit_price",
+                "max_unit_price",
+            ],
+        )
+        negotiations = pd.concat((negotiations, issues), axis=1)
+        if len(negotiations):
+            negotiations.drop("issues", axis=1, inplace=True)
+            for c in negotiations.columns:
+                if not c.endswith("step") and not c.endswith("steps"):
+                    continue
+                negotiations.loc[negotiations[c] < 0, c] = float("nan")
+                negotiations[c.replace("step", "relative_time")] = negotiations[c] / nsteps if nsteps else 0.0
+        offers = pd.DataFrame.from_records(offers_list)
+        if len(offers):
+            offers["receiver_type"] = offers.receiver.apply(do_map)
+            offers["sender_type"] = offers.sender.apply(do_map)
+            for c in offers.columns:
+                if not c.endswith("step") and not c.endswith("steps"):
+                    continue
+                offers.loc[offers[c] < 0, c] = float("nan")
+                offers[c.replace("step", "relative_time")] = offers[c] / nsteps if nsteps else 0.0
+
     contracts = contracts.loc[
         :,
         [
@@ -218,7 +407,7 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
         contracts.loc[contracts[c] < 0, c] = float("nan")
         contracts[c.replace("step", "relative_time")] = contracts[c] / nsteps if nsteps else 0.0
 
-    contract_stats = []
+    contract_stats, negotiation_stats = [], []
     non_step_cols = [
         "unit_price",
         "quantity",
@@ -230,11 +419,12 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
     ]
 
     def calc_qp(x):
+        c = len(x)
         s = (x["quantity"] * x["unit_price"]).sum()
         q = x["quantity"].sum()
         if q:
-            return q, s / q
-        return 0, 0.0
+            return q, c, s / q
+        return 0, 0, 0.0
 
     for step in range(nsteps):
         s = None
@@ -251,12 +441,12 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
             qp.columns = ["qp"]
             # breakpoint()
             qp = pd.DataFrame(qp["qp"].tolist(), index=qp.index)
-            qp = qp.rename(columns={0: base + "_quantity", 1: base + "_unit_price"})
+            qp = qp.rename(columns={0: base + "_quantity", 1: base + "_count", 2: base + "_unit_price"})
             # qp = qp.reset_index()
             qp["step"] = step
             qp.set_index(["step"], append=True, inplace=True)
             qp = qp.melt(
-                value_vars=[base + "_quantity", base + "_unit_price"],
+                value_vars=[base + "_quantity", base + "_count", base + "_unit_price"],
                 ignore_index=False,
             )
             if s is None:
@@ -286,9 +476,13 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
     for c in contract_stats.columns:
         if c.endswith("quantity"):
             contract_stats[c] = contract_stats[c].astype(int)
+        if c.endswith("count"):
+            contract_stats[c] = contract_stats[c].astype(int)
         if c.endswith("unit_price"):
             contract_stats[c] = contract_stats[c].astype(float)
 
+    # contract_stats.drop_duplicates(inplace=True, subset=["buyer", "seller", "step"])
+
     if nonzero(path / BREACHES_FILE):
         breaches = pd.read_csv(path / BREACHES_FILE, index_col=0)
     else:
@@ -354,7 +548,7 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
             if len(ns) > 1:
                 col_name = f"{ns[0]}_{aname}_{ns[-1]}"
             else:
-                col_name = f"{ns[0]}_{aname}"
+                col_name = f"{n}_{aname}"
             if col_name not in stats.columns:
                 continue
             colnames.append(col_name)
@@ -364,7 +558,7 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
             x["relative_time"] = x["step"] / nsteps
         else:
             x["relative_time"] = 0.0
-        x["agent"] = aname
+        x["name"] = aname
         x["world"] = wname
         x["tournament"] = tname
         for k, v in ainfo.items():
@@ -377,6 +571,7 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
             col_name = f"{n}_{p}"
             colnames.append(col_name)
         x = stats.loc[:, colnames].reset_index().rename(columns=dict(index="step"))
+        x.columns = ["_".join(_.split("_")[:-1]) if _.endswith(f"_{p}") else _ for _ in x.columns ]
         if len(x):
             x["relative_time"] = x["step"] / nsteps
         else:
@@ -400,21 +595,31 @@ def parse_world(path, tname, wname, nsteps, agents, w_indx, base_indx):
         world_info,
         contracts,
         contract_stats,
+        negotiations,
+        offers,
+        negotiation_stats,
         breaches,
     )
 
 
 def get_basic_world_info(path, tname):
-    stats = pd.read_csv(path / STATS_FILE, index_col=0).to_dict('list')
+    stats = pd.read_csv(path / STATS_FILE, index_col=0).to_dict("list")
     adata = json.load(open(path / AGENTS_JSON_FILE))
-    info = json.load(open(path / INFO_FILE))
-    worlds = [dict(name=path.name, tournament=tname, tournament_indx=0, path=path, n_steps=info["n_steps"])]
+    winfo = json.load(open(path / INFO_FILE))
+    worlds = [dict(name=path.name, tournament=tname, tournament_indx=0
+    , path=path, n_steps=winfo["n_steps"])]
     agents = []
     for i, (aname, info) in enumerate(adata.items()):
         if f"score_{aname}" not in stats.keys():
             continue
         score = stats[f"score_{aname}"][-1]
-        agents.append(dict(id=i, name=aname, world=worlds[0]["name"], tournament=tname, final_score=score, type=info["type"]))
+        aginfo = winfo["agent_profiles"][aname]
+        if "costs" in aginfo.keys():
+            aginfo["cost"] = float(np.asarray(aginfo["costs"]).min())
+            del aginfo["costs"]
+        dd = dict(id=i, name=aname, world=worlds[0]["name"], tournament=tname, final_score=score, type=info["type"])
+        dd = {**dd, **aginfo}
+        agents.append( dd)
     return worlds, pd.DataFrame.from_records(agents)
 
 
@@ -429,6 +634,7 @@ def get_data(base_folder):
         [],
     )
     contracts, contract_stats, breaches = [], [], []
+    negotiations, offers, neg_stats = [], [], []
     if is_tournament(base_folder):
         paths = get_torunaments(base_folder)
     elif is_world(base_folder):
@@ -436,6 +642,8 @@ def get_data(base_folder):
     else:
         raise ValueError(f"Folder {str(base_folder)} contains not tournament or world logs")
     for i, t in enumerate(paths):
+        indx = i + 1
+        base_indx = (i + 1) * 1_000_000
         if t is not None:
             print(f"Processing {t.name} [{i} of {len(paths)}]", flush=True)
             tournaments.append(dict(id=indx, path=t, name=t.name))
@@ -446,12 +654,10 @@ def get_data(base_folder):
             tname = "none"
             tournaments.append(dict(id=tname, path=base_folder.parent, name=tname))
             w, a = get_basic_world_info(base_folder, tname)
-        indx = i + 1
-        base_indx = (i + 1) * 1_000_000
         for j, world in enumerate(w):
             print(f"\tWorld {world['name']} [{j} of {len(w)}]", flush=True)
             wagents = a.loc[a.world == world["name"]]
-            ag, pr, wo, co, cs, br = parse_world(
+            ag, pr, wo, co, cs, ng, of, ns, br = parse_world(
                 Path(world["path"]),
                 tname,
                 world["name"],
@@ -460,24 +666,52 @@ def get_data(base_folder):
                 base_indx + j + 1,
                 base_indx + j + 1,
             )
-            agent_stats.append(ag)
-            product_stats.append(pr)
-            world_stats.append(wo)
-            contracts.append(co)
-            contract_stats.append(cs)
-            breaches.append(br)
-        worlds.append(pd.DataFrame.from_records(w))
-        agents.append(a)
+            if ag is not None and len(ag):
+                agent_stats.append(ag)
+            if pr is not None and len(pr):
+                product_stats.append(pr)
+            if wo is not None and len(wo):
+                world_stats.append(wo)
+            if co is not None and len(co):
+                contracts.append(co)
+            if cs is not None and len(cs):
+                contract_stats.append(cs)
+            if ng is not None and len(ng):
+                negotiations.append(ng)
+            if of is not None and len(of):
+                offers.append(of)
+            if ns is not None and len(ns):
+                neg_stats.append(ns)
+            if br is not None and len(br):
+                breaches.append(br)
+        if w is not None and len(w):
+            worlds.append(pd.DataFrame.from_records(w))
+        if a is not None and len(a):
+            agents.append(a)
 
     tournaments = pd.DataFrame.from_records(tournaments)
-    worlds = pd.concat(worlds, ignore_index=True)
-    agents = pd.concat(agents, ignore_index=True)
-    agent_stats = pd.concat(agent_stats, ignore_index=True)
-    product_stats = pd.concat(product_stats, ignore_index=True)
-    world_stats = pd.concat(world_stats, ignore_index=True)
-    contracts = pd.concat(contracts, ignore_index=True)
-    contract_stats = pd.concat(contract_stats, ignore_index=True)
-    breaches = pd.concat(breaches, ignore_index=True)
+    if worlds is not None and len(worlds):
+        worlds = pd.concat(worlds, ignore_index=True)
+    if agents is not None and len(agents):
+        agents = pd.concat(agents, ignore_index=True)
+    if agent_stats is not None and len(agent_stats):
+        agent_stats = pd.concat(agent_stats, ignore_index=True)
+    if product_stats is not None and len(product_stats):
+        product_stats = pd.concat(product_stats, ignore_index=True)
+    if world_stats is not None and len(world_stats):
+        world_stats = pd.concat(world_stats, ignore_index=True)
+    if contracts is not None and len(contracts):
+        contracts = pd.concat(contracts, ignore_index=True)
+    if contract_stats is not None and len(contract_stats):
+        contract_stats = pd.concat(contract_stats, ignore_index=True)
+    if negotiations is not None and len(negotiations):
+        negotiations = pd.concat(negotiations, ignore_index=True)
+    if offers is not None and len(offers):
+        offers = pd.concat(offers, ignore_index=True)
+    if neg_stats is not None and len(neg_stats):
+        neg_stats = pd.concat(neg_stats, ignore_index=True)
+    if breaches is not None and len(breaches):
+        breaches = pd.concat(breaches, ignore_index=True)
     return (
         tournaments,
         worlds,
@@ -487,6 +721,9 @@ def get_data(base_folder):
         world_stats,
         contracts,
         contract_stats,
+        negotiations,
+        offers,
+        neg_stats,
         breaches,
     )
 
@@ -504,6 +741,9 @@ def main(folder: Path, max_worlds: int):
         world_stats,
         contracts,
         contract_stats,
+        negotiations,
+        offers,
+        neg_stats,
         breaches,
     ) = get_data(folder)
     dst_folder = folder / VISDATA_FOLDER
@@ -518,6 +758,9 @@ def main(folder: Path, max_worlds: int):
             world_stats,
             contracts,
             contract_stats,
+            negotiations,
+            offers,
+            neg_stats,
             breaches,
         ),
         (
@@ -529,9 +772,14 @@ def main(folder: Path, max_worlds: int):
             "world_stats",
             "contracts",
             "contract_stats",
+            "negotiations",
+            "offers",
+            "neg_stats",
             "breaches",
         ),
     ):
+        if df is None or len(df) == 0:
+            continue
         df = adjust_type_names(df)
         df.to_csv(dst_folder / f"{name}.csv", index=False)
 
@@ -545,9 +793,6 @@ def has_visdata(folder: Path):
         "agent_stats",
         "product_stats",
         "world_stats",
-        "contracts",
-        "contract_stats",
-        "breaches",
     )
     if folder.name != VISDATA_FOLDER:
         folder /= VISDATA_FOLDER
